from __future__ import annotations

from dissect.ffs.exceptions import (
    Error,
    FileNotFoundError,
    NotADirectoryError,
    NotASymlinkError,
)
from dissect.ffs.ffs import FFS

__all__ = [
    "FFS",
    "Error",
    "FileNotFoundError",
    "NotADirectoryError",
    "NotASymlinkError",
]
